use crate::ffi::{SessionDesc, SlangMatrixLayoutMode};

#[cxx::bridge(namespace = "slang_compiler")]
mod ffi {
    #[namespace = ""]
    #[repr(i32)]
    enum SlangCompileTarget {
        SLANG_TARGET_UNKNOWN,
        SLANG_TARGET_NONE,
        SLANG_GLSL,
        SLANG_HLSL = 5,
        SLANG_SPIRV
    }

    // SlangProfileID is a combination of anonymous enum and preprocessor hack ultimately used
    // to generate switch statements for the various supported profiles - see `slang-profile-defs.h`
    // We'll define the names here and let C++ deal with mapping them to the ID
    #[namespace = ""]
    #[repr(u32)]
    enum SlangProfileID_rs {
        spirv_1_0,
    }

    // SlangTargetFlags is an anonymous enum in C++, use a shared enum (in the `slang_compiler` namespace)
    // here as a nicer API. Also means we need to manage the values by hand instead of relying on cxx.
    // https://stackoverflow.com/a/7147049
    #[repr(u32)]
    enum SlangTargetFlags {
        SLANG_TARGET_FLAG_PARAMETER_BLOCKS_USE_REGISTER_SPACES = 32,
        SLANG_TARGET_FLAG_GENERATE_WHOLE_PROGRAM = 256,
        SLANG_TARGET_FLAG_DUMP_IR = 512,
        SLANG_TARGET_FLAG_GENERATE_SPIRV_DIRECTLY = 1024,
    }

    #[namespace = ""]
    #[repr(u32)]
    enum SlangFloatingPointMode {
        SLANG_FLOATING_POINT_MODE_DEFAULT,
        SLANG_FLOATING_POINT_MODE_FAST,
        SLANG_FLOATING_POINT_MODE_PRECISE,
    }

    #[namespace = ""]
    #[repr(u32)]
    enum SlangLineDirectiveMode {
        SLANG_LINE_DIRECTIVE_MODE_DEFAULT,
        SLANG_LINE_DIRECTIVE_MODE_NONE,
        SLANG_LINE_DIRECTIVE_MODE_STANDARD,
        SLANG_LINE_DIRECTIVE_MODE_GLSL,
        SLANG_LINE_DIRECTIVE_MODE_SOURCE_MAP,
    }

    #[namespace = ""]
    #[repr(u32)]
    enum SlangMatrixLayoutMode {
        SLANG_MATRIX_LAYOUT_MODE_UNKNOWN,
        SLANG_MATRIX_LAYOUT_ROW_MAJOR,
        SLANG_MATRIX_LAYOUT_COLUMN_MAJOR,
    }

    // Some slang enums are defined in the root namespace
    #[namespace = ""]
    unsafe extern "C++" {
        include!("slang.h");
        type SlangCompileTarget;
        type SlangFloatingPointMode;
        type SlangLineDirectiveMode;
        type SlangMatrixLayoutMode;
    }

    #[namespace = "slang"]
    #[repr(u32)]
    enum CompilerOptionName {
        MacroDefine
    }

    #[namespace = "slang"]
    #[repr(u32)]
    enum CompilerOptionValueKind {
        Int,
        String,
    }

    #[namespace = "slang"]
    unsafe extern "C++" {
        include!("slang.h");
        type CompilerOptionName;
        type CompilerOptionValueKind;
    }

    #[derive(Clone)]
    struct CompilerOptionValue {
        kind: CompilerOptionValueKind,
        int_value_0: i32,
        int_value_1: i32,
        string_value_0: String,
        string_value_1: String,
    }

    #[derive(Clone)]
    struct CompilerOptionEntry {
        name: CompilerOptionName,
        value: CompilerOptionValue,
    }

    #[derive(Clone)]
    struct TargetDesc {
        format: SlangCompileTarget,
        profile: SlangProfileID_rs,
        flags: Vec<SlangTargetFlags>,
        floating_point_mode: SlangFloatingPointMode,
        line_directive_mode: SlangLineDirectiveMode,
        force_glsl_scalar_buffer_layout: bool,
        compiler_option_entries: Vec<CompilerOptionEntry>,
    }

    #[derive(Clone)]
    struct PreprocessorMacroDesc {
        name: String,
        value: String,
    }

    #[derive(Clone)]
    struct SessionDesc {
        targets: Vec<TargetDesc>,
        default_matrix_layout_mode: SlangMatrixLayoutMode,
        search_paths: Vec<String>,
        preprocessor_macros: Vec<PreprocessorMacroDesc>,
        enable_effect_annotations: bool,
        allow_glsl_syntax: bool,
        compiler_option_entries: Vec<CompilerOptionEntry>,
    }

    unsafe extern "C++" {
        include!("lib.h");

        type GlobalSession;
        fn create_global_session() -> SharedPtr<GlobalSession>;

        type Session;
        #[must_use]
        fn get_loaded_module_count(self: &Session) -> i64;

        #[must_use]
        fn create_session(session_desc: SessionDesc, global_session: SharedPtr<GlobalSession>) -> Result<UniquePtr<Session>>;
    }
}

impl Default for SessionDesc {
    fn default() -> Self {
        SessionDesc {
            targets: vec![],
            default_matrix_layout_mode: SlangMatrixLayoutMode::SLANG_MATRIX_LAYOUT_ROW_MAJOR,
            search_paths: vec![],
            preprocessor_macros: vec![],
            enable_effect_annotations: false,
            allow_glsl_syntax: false,
            compiler_option_entries: vec![],
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::ffi::{create_global_session, create_session};

    #[test]
    fn can_create_global_session() {
        let _global_session = create_global_session();
    }

    #[test]
    fn can_create_session() {
        let global_session = create_global_session();
        let session = create_session(Default::default(), global_session).expect("unable to create session");

        assert_eq!(session.get_loaded_module_count(), 0);
    }
}